/*
 * Usage (random): new_gen -random [-exact] -n=[n] -m=[m]
 * Generates random graph by repeatedly adding edges.
 * default: U(1, n) vertices, U(1, min(m, n * (n - 1) / 2)) edges
 * -exact: n vertices, m edges
 *
 * Usage (normal): new_gen -n=[number of vertices] -comps=[number of SCCs] [condensation DAG params] [SCCs params]
 * Alternative for -comps=n: -dag
 * Alternative for -comps=1: -scc
 *
 * DAG params:
 * -dag-edges=[m]: number of edges in DAG, otherwise U(0, min(2000, comps * (comps - 1) / 2))
 * -dag-hamiltonian: initially, add edges (i, i+1), then add edges up to m
 *
 * SCC params:
 * -scc-k=[k]: for passing a single parameter into any generation mode (I'm not parsing 10 different opts)
 *
 * *note*: function `merge_recursively` performs what I call merging, that is,
 * replacing i-th vertex of `base` with graph `subgraphs[i]`. this method is used
 * for generating much more complex components than simple random
 *
 * -scc-fractal: generate graph with k vertices, then merge it with itself until total size reaches n
 * -scc-fractal-random: same as -scc-fractal, but generate new graph with k vertices every time
 * -scc-random-2: generate base SCC with k vertices and smaller SCCs using simple random, the merge them
 * -scc-random-3: same as -scc-random-2, but smaller SCCs are generated as in -scc-random-2
 * -scc-hamiltonian: initially, add edges (i, i+1) and (n, 1), then add edges up to k
 * -scc-hamiltonian-short: same as -scc-hamiltonian, but only add edges that cover <log(n) cycle edges
 * default or -scc-from-random: generate random graph with 3n vertices s.t. its largest SCC has size n, return it
 *
 * *note*: all generation modes can be used without -scc-k while still generating good graphs
 *
 * *note*: some generation modes return graphs with <n vertices and possibly >2000 edges.
 * please exercise caution (and use validator)
 *
 * -shuffle: shuffle edges in output
 */

#include "testlib.h"
#include <bits/stdc++.h>

using namespace std;

class SccGraph {
private:
    vector<vector<int>> gr;
    vector<int> comp_color;

    vector<int> st;
    vector<int> t;
    int timer = 0;
    int cc = 0;

    void dfs(int v) {
        int tin = timer++;
        t[v] = tin;
        st.push_back(v);
        for (int to : gr[v]) {
            if (t[to] == -1) {
                dfs(to);
                t[v] = min(t[v], t[to]);
            } else if (comp_color[to] == -1) {
                t[v] = min(t[v], t[to]);
            }
        }
        if (tin == t[v]) {
            while (st.back() != v) {
                comp_color[st.back()] = cc;
                st.pop_back();
            }
            comp_color[v] = cc;
            st.pop_back();
            cc++;
        }
    }

public:
    int n;
    SccGraph() : n(1), gr(1) { }
    explicit SccGraph(int n) : n(n), gr(n) { }

    void add_edge(int u, int v) {
        gr[u].push_back(v);
    }

    pair<int, vector<int>> get_scc() {
        st.clear();
        timer = 0;
        cc = 0;
        t.assign(n, -1);
        comp_color.assign(n, -1);
        for (int i = 0; i < n; ++i) {
            if (t[i] == -1) {
                dfs(i);
            }
        }
        return {cc, comp_color};
    }

    int edges_cnt() const {
        int res = 0;
        for (int i = 0; i < n; ++i) {
            res += gr[i].size();
        }
        return res;
    }

    vector<pair<int, int>> edges() const {
        vector<pair<int, int>> res;
        for (int i = 0; i < n; ++i) {
            for (int to : gr[i]) {
                res.emplace_back(i, to);
            }
        }
        return res;
    }

    void output() const {
        auto e = edges();
        if (opt<bool>("shuffle")) {
            shuffle(e.begin(), e.end());
        }
        printf("%d %d %d\n", n, (int)e.size(), opt<int>("g"));
        for (auto [u, v] : e) {
            printf("%d %d\n", u + 1, v + 1);
        }
    }
};

SccGraph generate_scc_from_random(int n, int m, int max_n, int max_m) {
    set<pair<int, int>> edges;
    vector<pair<int, int>> edges_v;

    for (int i = 0; i < m; ++i) {
        while (true) {
            int u = rnd.next(0, n - 1);
            int v = rnd.next(0, n - 2);
            if (v >= u) v++;
            if (!edges.count({u, v}) && !edges.count({v, u})) {
                edges.emplace(u, v);
                edges_v.emplace_back(u, v);
                break;
            }
        }
    }

    auto get_largest_scc = [&](int edges_to_take) {
        SccGraph cur_gr(n);
        for (int i = 0; i < edges_to_take; ++i) {
            auto [u, v] = edges_v[i];
            cur_gr.add_edge(u, v);
        }

        auto [cc, comp_color] = cur_gr.get_scc();

        vector<int> comp_sizes(cc);
        for (int i = 0; i < n; ++i) {
            comp_sizes[comp_color[i]]++;
        }

        int mx = 0;
        for (int i = 0; i < cc; ++i) {
            if (comp_sizes[i] > comp_sizes[mx]) {
                mx = i;
            }
        }

        vector<int> renum(n, -1);
        int cur = 0;
        for (int i = 0; i < n; ++i) {
            if (comp_color[i] == mx) {
                renum[i] = cur++;
            }
        }

        SccGraph largest_scc(comp_sizes[mx]);

        for (int i = 0; i < edges_to_take; ++i) {
            auto [u, v] = edges_v[i];
            if (comp_color[u] == mx && comp_color[v] == mx) {
                largest_scc.add_edge(renum[u], renum[v]);
            }
        }

        return largest_scc;
    };

    int tl = 0, tr = m + 1;
    while (tr - tl > 1) {
        int tm = (tl + tr) / 2;
        auto largest_scc = get_largest_scc(tm);
        if (largest_scc.n <= max_n && largest_scc.edges_cnt() <= max_m) {
            tl = tm;
        } else {
            tr = tm;
        }
    }

    return get_largest_scc(tl);
}

void choose_more_edges(int n, int m, set<pair<int, int>>& edges, bool only_right = false) {
    vector<pair<int, int>> available;
    for (int u = 0; u < n; ++u) {
        for (int v = u + 1; v < n; ++v) {
            if (!edges.count({u, v}) && !edges.count({v, u})) {
                available.emplace_back(u, v);
            }
        }
    }
//    ensuref(available.size() >= m - (int)edges.size(), "too many edges");
    vector<int> idx = rnd.distinct(min(m - (int)edges.size(), (int)available.size()), 0, (int)available.size() - 1);
    for (int i : idx) {
        auto [u, v] = available[i];
        if (only_right || rnd.next(0, 1)) {
            edges.emplace(u, v);
        } else {
            edges.emplace(v, u);
        }
    }
}

SccGraph generate_scc_hamiltonian(int n, int m) {
    set<pair<int, int>> edges;
    for (int i = 0; i < n - 1; ++i) {
        edges.emplace(i, i + 1);
    }
    edges.emplace(n - 1, 0);
    choose_more_edges(n, m, edges);
    SccGraph result(n);
    for (auto [u, v] : edges) {
        result.add_edge(u, v);
    }
    return result;
}

SccGraph generate_scc_hamiltonian_with_short_edges(int n, int m) {
    set<pair<int, int>> edges;
    for (int i = 0; i < n - 1; ++i) {
        edges.emplace(i, i + 1);
    }
    edges.emplace(n - 1, 0);
    int threshold = (int)log(n);
    for (int i = n; i < m; ++i) {
        while (true) {
            int u = rnd.next(0, n - 1);
            int v = (u + rnd.next(-threshold, threshold) + n) % n;
            if (u != v && !edges.count({u, v}) && !edges.count({v, u})) {
                edges.emplace(u, v);
                break;
            }
        }
    }
    SccGraph result(n);
    for (auto [u, v] : edges) {
        result.add_edge(u, v);
    }
    return result;
}

SccGraph generate_scc_with_blocks(int n, int blocks = -1) {
    if (blocks == -1) {
        blocks = rnd.next(1, n - 2);
    }
    set<pair<int, int>> edges;
    vector<int> idx = rnd.distinct(blocks - 1, 3, n - 1);
    sort(idx.begin(), idx.end());
    idx.push_back(n);
    vector<int> perm = rnd.perm(n);
    for (int i = 0; i < idx[0] - 1; ++i) {
        edges.emplace(perm[i], perm[i + 1]);
    }
    edges.emplace(perm[idx[0] - 1], perm[0]);
    for (int b = 0; b < blocks - 1; ++b) {
        int from, to;
        if (idx[b + 1] - idx[b] > 1) {
            from = rnd.next(0, idx[b] - 1);
            to = rnd.next(0, idx[b] - 1);
        } else {
            auto ends = rnd.distinct(2, 0, idx[b] - 1);
            from = ends[0];
            to = ends[1];
        }
        edges.emplace(perm[from], perm[idx[b]]);
        for (int i = idx[b]; i < idx[b + 1] - 1; ++i) {
            edges.emplace(perm[i], perm[i + 1]);
        }
        edges.emplace(perm[idx[b + 1] - 1], perm[to]);
    }
    SccGraph result(n);
    for (auto [u, v] : edges) {
        result.add_edge(u, v);
    }
    return result;
}

SccGraph merge_recursively(const SccGraph& base, const vector<SccGraph>& subgraphs) {
    int k = subgraphs.size();
    vector<int> head(k);
    int size = 0;
    for (int i = 0; i < k; ++i) {
        head[i] = size;
        size += subgraphs[i].n;
    }
    SccGraph result(size);
    for (int i = 0; i < k; ++i) {
        for (auto [u, v] : subgraphs[i].edges()) {
            result.add_edge(head[i] + u, head[i] + v);
        }
    }
    for (auto [u, v] : base.edges()) {
        int from = rnd.next(head[u], head[u] + subgraphs[u].n - 1);
        int to = rnd.next(head[v], head[v] + subgraphs[v].n - 1);
        result.add_edge(from, to);
    }
    return result;
}

SccGraph generate_scc_two_steps(int n, int k = -1) {
    if (3 > n / 3) {
        return generate_scc_with_blocks(n);
    }
    if (k == -1) {
        k = rnd.next(3, n / 3);
    }
    vector<int> subgraph_sizes = rnd.partition(k, n, 3);
    SccGraph base = generate_scc_with_blocks(k, -1);
    vector<SccGraph> subgraphs(k);
    for (int i = 0; i < k; ++i) {
        int s = subgraph_sizes[i];
        subgraphs[i] = generate_scc_with_blocks(s, -1);
    }
    return merge_recursively(base, subgraphs);
}

SccGraph generate_scc_three_steps(int n, int k = -1) {
    if (3 > n / 3) {
        return generate_scc_with_blocks(n);
    }
    if (k == -1) {
        k = rnd.next(3, n / 3);
    }
    vector<int> subgraph_sizes = rnd.partition(k, n, 3);
    SccGraph base = generate_scc_with_blocks(k, -1);
    vector<SccGraph> subgraphs(k);
    for (int i = 0; i < k; ++i) {
        int s = subgraph_sizes[i];
        subgraphs[i] = generate_scc_two_steps(s, -1);
    }
    return merge_recursively(base, subgraphs);
}

SccGraph generate_scc_fractal(int n, const SccGraph& base) {
    SccGraph gr = base;
    while (base.n * gr.n <= n) {
        gr = merge_recursively(base, {(size_t)base.n, gr});
    }
    return gr;
}

SccGraph generate_scc_random_fractal(int n, int k = -1) {
    SccGraph gr = generate_scc_with_blocks(k != -1 ? k : rnd.next(3, 6));
    while (gr.n * (k != -1 ? k : 6) <= n) {
        vector<SccGraph> subgraphs(gr.n);
        for (int i = 0; i < gr.n; ++i) {
            subgraphs[i] = generate_scc_with_blocks(k != -1 ? k : rnd.next(3, 6));
        }
        gr = merge_recursively(gr, subgraphs);
    }
    return gr;
}

void validate_scc(SccGraph gr) {
    auto [cc, comp_color] = gr.get_scc();
    ensuref(cc == 1, "graph has to be strongly connected");
}

SccGraph generate_scc_from_opts(int n) {
    if (n == 1) return {};
    SccGraph result;

    int k = opt<int>("scc-k", -1);

    if (opt<bool>("scc-fractal")) {
        if (k == -1) {
            k = rnd.next(3, 6);
        }
        SccGraph base = generate_scc_with_blocks(k);
        result = generate_scc_fractal(n, base);

    } else if (opt<bool>("scc-fractal-random")) {
        result = generate_scc_random_fractal(n, k);
    } else if (opt<bool>("scc-random-2")) {
        result = generate_scc_two_steps(n, k);
    } else if (opt<bool>("scc-random-3")) {
        result = generate_scc_three_steps(n, k);
    } else if (opt<bool>("scc-hamiltonian")) {
        if (k == -1) {
            k = n;
        }
        result = generate_scc_hamiltonian(n, k);
    } else if (opt<bool>("scc-hamiltonian-short")) {
        if (k == -1) {
            k = n;
        }
        result = generate_scc_hamiltonian_with_short_edges(n, k);
    } else /* if (opt<bool>("scc-from-random")) */ {
        bool testlib_screw_you = opt<bool>("scc-from-random");
        result = generate_scc_from_random(n * 3, n * 6, n, 2000);
    }

    validate_scc(result);

    return result;
}

SccGraph generate_dag_from_opts(int comps) {
    SccGraph dag(comps);
    if (comps > 1) {
        set<pair<int, int>> edges;
        int dag_edges;
        if (opt<bool>("dag-hamiltonian")) {
            dag_edges = opt<int>("dag-edges", rnd.next(comps - 1, min(2000, comps * (comps - 1) / 2)));
            for (int i = 0; i < comps - 1; ++i) {
                edges.emplace(i, i + 1);
            }
        } else {
            dag_edges = opt<int>("dag-edges", rnd.next(0, min(2000, comps * (comps - 1) / 2)));
        }
        choose_more_edges(comps, dag_edges, edges, true);
        for (auto [u, v] : edges) {
            dag.add_edge(u, v);
        }
    }
    return dag;
}

SccGraph generate_random_from_opts() {
    int n, m;
    bool chain = opt<bool>("hamiltonian");
    if (!opt<bool>("exact")) {
        int max_n = opt<int>("n");
        int max_m = opt<int>("m");
        n = rnd.next(2, max_n);
        m = rnd.next(chain ? n - 1 : 1, min(n * (n - 1) / 2, max_m));
    } else {
        n = opt<int>("n");
        m = opt<int>("m");
    }
    set<pair<int, int>> edges;
    if (chain) {
        for (int i = 0; i < n - 1; ++i) {
            edges.emplace(i, i + 1);
        }
    }
    choose_more_edges(n, m, edges);
    SccGraph result(n);
    for (auto [u, v] : edges) {
        result.add_edge(u, v);
    }
    return result;
}

int main(int argc, char *argv[]) {
    registerGen(argc, argv, 1);

    // this is as hacky as it gets, deal with it
    rnd.setSeed(argc - 1, argv + 1);

    if (has_opt("seed")) {
        int seed = opt<int>("seed", 0);
        rnd.setSeed(seed);
    }

    if (opt<bool>("random")) {
        generate_random_from_opts().output();
        return 0;
    }

    int n = opt<int>("n");
    int comps = opt<int>("comps", -1);

    if (opt<bool>("dag")) {
        comps = n;
    }
    if (opt<bool>("scc")) {
        comps = 1;
    }
    if (comps == -1) {
        comps = rnd.next(1, n);
    }

    SccGraph dag = generate_dag_from_opts(comps);

    vector<SccGraph> scc(comps);
    {
        int one_comps = clamp((3 * comps - n + 1) / 2, 0, comps);
        auto non_one_comps = rnd.distinct(comps - one_comps, 0, comps - 1);
        auto scc_sizes = rnd.partition(comps - one_comps, n - one_comps, 3);
        for (int i = 0; i < non_one_comps.size(); ++i) {
            scc[non_one_comps[i]] = generate_scc_from_opts(scc_sizes[i]);
        }
    }

    merge_recursively(dag, scc).output();
}
